# slope

slope is an interpreted s-expression based programming language. The interpreter is implemented in the [go programming language](https://go.dev). Those familiar with scheme or lisp will likely find themselves on reasonably familiar ground. Given its roots in golang, it wraps a lot of the go standard library in convenient ways for interpreted/script based programming. The [REPL](https://en.wikipedia.org/wiki/Read%E2%80%93eval%E2%80%93print_loop) comes with command completion, readline-style line editing, and procedure help/definitions (via the `usage` procedure).


**Table of Contents**
1. [Why another programming language?](#why-another-programming-language)
2. [Acknowledgements](#acknowledgements)
3. [The Language](#the-language)
   1. [Types](#types)
   2. [Special Forms](#special-forms)
   3. [Library](#library)
       1. [Constant Values](#values)
       2. [Numbers](#number-related)
       3. [Conditionals](#conditional)
       4. [Lists](#list-related)
       5. [Strings](#string-related)
       6. [IO](#io-related)
       7. [System](#system-related)
       8. [OS](#os-related)
       9. [Net](#net-related)
       10. [Date/Time](#date-time-related)
       11. [GUI (Optional)](#gui-related)
4. [Preload](#preload)
5. [Modules](#modules)
   1. [Package Management](#package-management)
   2. [How Modules Are Processed](#how-modules-are-processed)
   3. [Module Documentation](#module-documentation)
   4. [Module Files](#module-files)
       1. [main.slo](#main-slo)
       2. [Additional Source Files](#additional-source-files)
       3. [module.json](#module-json)
6. [Building From Source](#building-from-source)
7. [Running](#running)
   1. [The REPL](#the-repl)
   2. [One Liners](#one-liners)
   3. [Slope Files](#slope-files)
   4. [Other Options](#other-options)
8. [License](#license)


## Why another programming language?

Like many many projects on the internet: to learn and have fun. I have definitely learned different things than my [previous language project](https://git.rawtext.club/sloum/nimf) while working on this codebase, and I think have ended up with a much more realistically usable language (though your mileage may vary).


## Acknowledgements

This interpreter is based on some [existing code](https://gist.github.com/pkelchte/c2bd76b9f8f9cd603b3c). I rewrote sections of the lexer and parser (areas I enjoy working in), left the apply and environment setup alone, and made a few small changes to the eval function. As such, this could be considered a fork of that code with some added core language constructs as well as a significantly expanded library (another area I enjoy working in).


## The language



### Types

slope recognizes a few types, that are generally created and used dynamically:

- `number`
  - implemented as `float64` under the hood
  - hexidecimal can be written with a leading `0x`, as in `0xFFF`
  - octal can be written with a leading `0`, as in `077`
  - in strings, values can be escaped to characters as decimal, octal, or hex: `"\27[1mI am bold text\033[0m, I am not bold. I am a hex based char: \0x84"`
- `symbol`
  - strings that represent language objects/constructs/vars
  - can be created using `(quote my-symbol)` or `'my-symbol `
- `bool`
  - as `#t` and `#f`
  - anything can be cast to bool with a `#t` value, except `#f` (the only truly falsy value without using `~bool`, see below)
- `string`
  - anything in "quotes"
  - allows for backslash escapes within the quotes (for chars by number or shortcuts to tab, newline, etc)
- `list`/`pair`
  - includes the empty list (null)
  - lists are the primary data structure
- `exception`
  - an error response
  - can be created using `(! "My exception text")`
  - can be tested for with `(exception? some-var)`
  - two functions exist to set behavior for exceptions:
    - `(exception-mode-panic)`, will panic and halt execution when an exception is encountered (_default_)
    - `(exception-mode-pass)`, will return the exception like any other value and it can be handled
    - the two modes can be switched on and off at various points in the code
- `io-handle`
    - represents files (in read or write mode), network connections, and string buffers
    - passed by _reference_ rather than by value (unlike all other slope types)

If the _optional_ gui module is installed, the following become available:

- `gui-root`
    - The application root of a gui
    - Holds all windows and is used as the reference point for many gui based procedures
    - Passed by _value_ like most slope types
- `gui-container`
    - an organizational structure used for laying out widgets
    - does not specify the type of container, just that the value is a `gui-container`
- `gui-widget`
    - the interface building block of gui applications
    - does not specify the type of widget, just that the value is a `gui-widget`



### Special Forms

There are a number of special forms in the language that will allow, for example, un-initialized symbols to be passed (for example, in `define`), among other things less available from regular lambdas (which are themselves a special form).

- `quote`
  - Can be shorthanded so that `'(1 2 3)` is equal to `(quote (1 2 3))`
- `if`
- `and`
- `or`
- `cond`
- `set!`
- `define`
  - Lexically scoped, such that its use from within a lambda scopes any defined variables to the scope of that lambda and its child scopes
- `lambda`
  - The body of a lambda has an implied `(begin ...)` surrounding it. As such, you may place non-nested procedure calls or values inside of a lambda body and the last will be returned
  - Variadic lambdas can be created by using the special param name `args-list` (which will be a list of all of the remaining arguments when accessed rom the lambda body). This param should be the last param in the lambda definition, as it will eat all arguments that come after its position. Any params defined after `args-list` have undefined behavior (they will generally not be created)
- `begin`
- `begin0`
  - Like begin, but returns the value resulting from the evaluation of the first expression. No shorthand exists for `begin0`
- `filter`
  - Takes a procedure that, in turn, takes a single argument. Each value from the second argument, a list, will be passed to the procedure. Anything truthy results are returned in a new list
- `load`
  - Loads and executes a slope file. Absolute path to file must be used
- `load-mod`
  - Loads a module. Module _name_ should be used
- `load-mod-file`
  - Used within modules to load additional files
- `usage`
  - Used to display the procedure signature and any additional information about the procedure, mostly used interactively
- `exists?`
  - Checks if the current environment is aware of the given symbol
- `eval`
  - Executes representations of slope code

While not a special form from a compiler implementation standpoint `list` has some syntactic sugar that is worthy of note. These four lines will all produce the same list:

``` slope
(list 1 2 3)
[1 2 3]
(quote 1 2 3)
'(1 2 3)
```

Quote and list both have some syntactic sugar to create a shorthand for their usage (`'()` for quote and `[ ... ]` for lists).



<details>
  <summary>Special Forms API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(quote [expression...])</code>: <code>bool</code></li>
    <li><code>(if [condition check: expression] [true branch: expression] [[false branch: expression]])</code>: <code>value</code></li>
    <li><code>(and [expression...])</code>: <code>bool</code></li>
    <li><code>(or [expression...])</code>: <code>bool</code></li>
    <li><code>(cond [([condition check: expression] [expression])...])</code>: <code>value</code></li>
    <li><code>(set! [variable name: symbol] [value|expression])</code>: <code>value/expression</code></li>
    <li><code>(define [variable name: symbol] [value|expression|procedure])</code>: <code>value/expression</code></li>
    <li><code>(lambda [([[argument: symbol...]])] [[expression...]])</code>: <code>procedure</code></li>
    <li><code>(begin [expression...])</code>: <code>value</code></li>
    <li><code>(begin0 [expression...])</code>: <code>value</code></li>
    <li><code>(filter [test: procedure] [list])</code>: <code>list</code></li>
    <li><code>(load [filepath: string...])</code>: <code>symbol</code></li>
    <li><code>(load-mod [filepath: string...])</code>: <code>symbol</code></li>
    <li><code>(exists? [symbol...])</code>: <code>bool</code></li>
    <li><code>(load-mod-file [filepath: string...])</code>: <code>symbol</code></li>
    <li><code>(usage [[module|builtin-procedure: string]] [[module-procedure: string|#f]])</code>: <code>()</code> Will display usage information for the given procedure or runtime value. If no argument is given <code>usage</code> will list known procedures. If a module name and `#f` are passed a list of known procedures for that modules will be printed. If a module name and a known procedure for that module are given then information about the procedure will be printed.</li>
    <li><code>(eval [expression|value] [[evaluate-string-as-code: bool]])</code>: <code>value</code></li>
  </ul>
</details>

### Library

#### Values

`PI`, `E`, `PHI`, `sys-args`, `stdin`, `stdout`, `stderr`, `devnull`


`sys-args` is populated with the arguments run at the command line, similar to C's `argv`

`stdin`, `stdout`, `stderr`, and `devnull` are set at runtime to an open IOHandle representing each IO file. This makes it easy to use them the same as you would any other file (by passing the IOHandle to the reader/writer).


#### Number related

Implemented:

`positive?`, `negative?`, `zero?`, `abs`, `floor`, `ceil`, `round`, `+`, `-`, `*`, `/`, `min`, `max`, `number->string`, `rune->string`, `rand`

Not implemented, but on my radar:

`modulo`, `remainder`, `quotient`, `expt`, `sin`, `cos`, `tan`, `atan`, `sqrt`

<details>
  <summary>Number API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(positive? [number])</code>: <code>bool</code></li>
    <li><code>(negative? [number])</code>: <code>bool</code></li>
    <li><code>(zero? [number])</code>: <code>bool</code></li>
    <li><code>(abs [number])</code>: <code>number</code></li>
    <li><code>(floor [number])</code>: <code>number</code></li>
    <li><code>(ceil [number])</code>: <code>number</code></li>
    <li><code>(round [number] [[number: decimal places to round to]])</code>: <code>number</code></li>
    <li><code>(+ [number...])</code>: <code>number</code></li>
    <li><code>(- [number...])</code>: <code>number</code></li>
    <li><code>(* [number...])</code>: <code>number</code></li>
    <li><code>(/ [number...])</code>: <code>number</code></li>
    <li><code>(min [number...])</code>: <code>number</code></li>
    <li><code>(max [number...])</code>: <code>number</code></li>
    <li><code>(number->string [number] [[base: number]])</code>: <code>string</code></li>
    <li><code>(rune->string [number])</code>: <code>string</code></li>
    <li><code>(rand [[max]] [[min]])</code>: <code>number</code>
  </ul>
</details>


#### Conditional

Implemented:

`>`, `>=`, `<`, `<=`, `not`, `equal?`, `number?`, `string?`, `bool?`, `symbol?`, `pair?`, `null?`, `list?`, `procedure?`, `atom?`, `~bool`, `assoc?`, `string-buf?`, `io-handle?`

`~bool` is not common in scheme-like languages and will cast any value as a boolean value using a looser set of rules. Under normal conditional rules in _slope_ anythong other than `#f` is considered truthy (`#t`). When using `~bool` the number `0` is falsy, the empty list `()` is falsy, the empty string `""` is falsy, and a closed IOHandle is falsy. Everything else, except the literal value `#f`, is true (`#t`).

So...

``` slope
> (if (~bool 0) "Truthy" "Falsy")
#1=> Falsy
> (if 0 "Truthy" "Falsy")
#2=> Truthy
```

<details>
  <summary>Conditional API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(> [number] [number])</code>: <code>bool</code></li>
    <li><code>(>= [number] [number])</code>: <code>bool</code></li>
    <li><code>(< [number] [number])</code>: <code>bool</code></li>
    <li><code>(<= [number] [number])</code>: <code>bool</code></li>
    <li><code>(not [value])</code>: <code>bool</code></li>
    <li><code>(equal? [value...])</code>: <code>bool</code></li>
    <li><code>(number? [value])</code>: <code>bool</code></li>
    <li><code>(string? [value])</code>: <code>bool</code></li>
    <li><code>(bool? [value])</code>: <code>bool</code></li>
    <li><code>(symbol? [value])</code>: <code>bool</code></li>
    <li><code>(pair? [value])</code>: <code>bool</code></li>
    <li><code>(list? [value])</code>: <code>bool</code></li>
    <li><code>(null? [value])</code>: <code>bool</code></li>
    <li><code>(procedure? [value])</code>: <code>bool</code></li>
    <li><code>(atom? [value])</code>: <code>bool</code></li>
    <li><code>(assoc? [value])</code>: <code>bool</code></li>
    <li><code>(io-handle? [value])</code>: <code>bool</code></li>
    <li><code>(string-buf? [value])</code>: <code>bool</code></li>
    <li><code>(~bool [value])</code>: <code>bool</code></li>
  </ul>
</details>

#### List related

Implemented:

`length`, `cons`, `car`, `cdr`, `append`, `list`, `map`, `for-each`, `list->string`, `ref`, `list-sort`, `reverse`, `assoc`, `member?`, `slice`, `list-seed`, `list-join`, `reduce`

- `list` can be shorthanded by using square brackets: `[1 2 3]` vs `(list 1 2 3)`
- `length` and `reverse` will take a string or a list.
- `list-sort` will sort a list in ascending order by casting all elements within the list as strings and doing a compare. The exception to this is numbers, which will be compared among themselves as numbers. In a mixed list numbers will come after other values. To sort descending, pass the results of `list-sort` to `reverse`

<details>
  <summary>List API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(length [list|string|exception|IOHandle:string-buf])</code>: <code>number</code></li>
    <li><code>(list [value...])</code>: <code>list</code></li>
    <li><code>(cons [value] [list])</code>: <code>list</code></li>
    <li><code>(car [list])</code>: <code>list</code></li>
    <li><code>(cdr [list])</code>: <code>list</code></li>
    <li><code>(append [list|value] [[value...])</code>: <code>list|string</code></li>
    <li><code>(map [procedure] [list...])</code>: <code>list</code></li>
    <li><code>(for-each [procedure] [list...])</code>: <code>list</code></li>
    <li><code>(list-join [list...])</code>: <code>list</code></li>
    <li><code>(list->string [list] [join-on: string])</code>: <code>string</code></li>
    <li><code>(ref [list|string] [index: number] [[set: value]])</code>: <code>value|list|string</code></li>
    <li><code>(list-sort [list] [[sub-list-index: number]])</code>: <code>list</code></li>
    <li><code>(reduce [procedure] [accumulator: value] [list])</code>: <code>value</code></li>
    <li><code>(reverse [list|string])</code>: <code>list</code></li>
    <li><code>(slice [list|string] [number] [[number]])</code>: <code>list|string</code></li>
    <li><code>(assoc [assoc-list] [key: value] [[value: value]])</code>: <code>value</code></li>
    <li><code>(member? [list] [value])</code>: <code>bool</code></li>
    <li><code>(list-seed [length: number] [value])</code>: <code>list</code> More efficient than recursion for building large lists and wont run into stack overflows</li>
  </ul>
</details>

#### String related

Implemented:

`length`, `slice`, `string->number`, `string-format`, `string->list`, `string-trim-space`, `string-index-of`, `ref`, `string-upper`, `string-lower`, `string-make-buf`, `string-buf-clear`, `regex-match?`, `regex-find`, `regex-replace`, `string-fields`, `string->md5`, `string->sha256`

- `length` will take a string, an IOHandle representing a string-buf, or a list. In the case of a string it will return the number of _runes_ as thought of in golang. Loosely equivalent to characters, rather than bytes.
- Standard read and write procedures work with the IOHandle returned from `string-make-buf`
- `length` and `reverse` will take a string or a list.
- `string-format` works like a simplified sprintf. All values are represented by `%v` (for value). You can pad values with spaces on the right by adding the desired field width after the percent: `%10v` with the string `"hello"` would produce `"     hello"` and you can left justify the text with a `-` after the percent: `%-10v` for the same variable would produce `"hello     "`. There are no other modifiers.


<details>
  <summary>String API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(length [list|string|exception|IOHandle:string-buf])</code>: <code>number</code></li>
    <li><code>(slice [list|string] [number] [[number]])</code>: <code>list|string</code></li>
    <li><code>(append [list|value] [[value...]])</code>: <code>list|string</code></li>
    <li><code>(string-format [format: string] [value...])</code>: <code>string</code> (replaces `{}`'s in string with successive values)</li>
    <li><code>(string->number [string] [[base: number]])</code>: <code>number/#f</code> If a non-decimal base is supplied, the input string will be parsed as an integer and any flaoting point value will be floored. A valid base passed with a string that does not parse to a number will return <code>#f</code></li>
    <li><code>(string->list [string] [[split-point: value]] [[count: number]])</code>: <code>list</code> (splits the first string at each instance of value, cast as a string)</li>
    <li><code>(string->md5 [string])</code>: <code>string</code></li>
    <li><code>(string->rune [string])</code>: <code>number</code></li>
    <li><code>(string->sha256 [string])</code>: <code>string</code></li>
    <li><code>(string-fields [string])</code>: <code>list</code></li>
    <li><code>(string-trim-space [string])</code>: <code>string</code></li>
    <li><code>(string-index-of [string] [string])</code>: <code>number</code> (`-1` if the string could not be found)</li>
    <li><code>(ref [list|string] [number])</code>: <code>value|list|string</code></li>
    <li><code>(string-upper [string])</code>: <code>string</code></li>
    <li><code>(string-lower [string])</code>: <code>string</code></li>
    <li><code>(string-make-buf)</code>: <code>IOHandle:string-buf</code></li>
    <li><code>(string-buf-clear [IOHandle:string-buf])</code>: <code>()</code></li>
    <li><code>(regex-match? [string (pattern)] [string])</code>: <code>bool</code></li>
    <li><code>(regex-find [string (pattern)] [string])</code>: <code>list</code></li>
    <li><code>(regex-replace [string (pattern)] [string])</code>: <code>string</code></li>
  </ul>
</details>

#### IO Related

Implemented:

`newline`, `display`, `display-lines`, `write`, `write-raw`, `read-line`, `read-char`, `read-all`, `read-all-lines`, `close` `file-create`, `file-create-temp`, `file-open-read`, `file-open-write`, `file-append-to`, `file-stat`, `file-name`

Not implemented, but on my radar:

`file-seek`


<details>
  <summary>IO API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(newline)</code>: <code>()</code></li>
    <li><code>(display [value...])</code>: <code>()</code></li>
    <li><code>(display-lines [value...])</code>: <code>()</code></li>
    <li><code>(write [string] [[IOHandle]])</code>: <code>()|IOHandle</code></li>
    <li><code>(write-raw [string] [[IOHandle]])</code>: <code>()|IOHandle</code></li>
    <li><code>(read-all [[IOHandle]])</code>: <code>string</code></li>
    <li><code>(read-line [[IOHandle]])</code>: <code>string|#f</code></li>
    <li><code>(read-char [[IOHandle]])</code>: <code>string|#f</code></li>
    <li><code>(read-all-lines [[IOHandle]])</code>: <code>list</code></li>
    <li><code>(close [IOHandle])</code>: <code>()</code></li>
    <li><code>(file-create [filepath: string])</code>: <code>IOHandle</code></li>
    <li><code>(file-create-temp [filename-pattern: string])</code>: <code>IOHandle</code></li>
    <li><code>(file-open-read [filepath: string])</code>: <code>IOHandle</code></li>
    <li><code>(file-open-write [filepath: string])</code>: <code>IOHandle</code></li>
    <li><code>(file-append-to [filepath: string] [string...])</code>: <code>()</code></li>
    <li><code>(file-name [IOHandle: file])</code>: <code>string</code></li>
    <li><code>(file-stat [filepath: string])</code>: <code>assoc-list|#f</code>
      <p>Will return <code>#f</code> if the file does not exist, but the procedure otherwise runs successfully. Any other errors accessing a file will raise an exception.</p>
      <p>The associative list will have the following keys: </p>
        <ul>
          <li><code>name</code></li>
          <li><code>size</code></li>
          <li><code>mode</code></li>
          <li><code>mod-time</code></li>
          <li><code>is-dir?</code></li>
          <li><code>is-symlink?</code></li>
          <li><code>path</code></li>
        </ul>
      <p><code>mode</code> will be in decimal as a number and can be converted to an octal string with <code>number->string</code>. <code>size</code> is in bytes. <code>path</code> will be an absolute path after following a symlink, or the path as provided to stat if the file is not a symlink.</p>
    </li>
  </ul>
</details>

#### System Related

Implemented:

`exit`, `license`, `apply`, `!`, `exception-mode-panic`, `exception-mode-pass`, `term-size`, `term-restore`, `term-char-mode`, `term-raw-mode`, `term-cooked-mode`, `term-sane-mode`, `signal-catch-sigint`, `exception-mode-pass?`, `exception-mode-panic?`, `mod-path`

- `(license)` will print the license terms for slope, this is mostly input as a convenience in the repl.
- The various term modes are an experimental feature at the moment, but should work as described


<details>
  <summary>System API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(exit [[number]])</code>: <code>()</code></li>
    <li><code>(license)</code>: <code>()</code></li>
    <li><code>(apply [procedure] [arguments: list])</code>: <code>value</code></li>
    <li><code>(! [exception-text: string])</code>: <code>exception</code></li>
    <li><code>(exception-mode-panic)</code>: <code>()</code></li>
    <li><code>(exception-mode-pass)</code>: <code>()</code></li>
    <li><code>(exception-mode-panic?)</code>: <code>bool</code></li>
    <li><code>(exception-mode-pass?)</code>: <code>bool</code></li>
    <li><code>(term-size)</code>: <code>list</code></li>
    <li><code>(term-restore)</code>: <code>()</code></li>
    <li><code>(term-char-mode)</code>: <code>()</code></li>
    <li><code>(term-raw-mode)</code>: <code>()</code></li>
    <li><code>(term-cooked-mode)</code>: <code>()</code></li>
    <li><code>(term-sane-mode)</code>: <code>()</code></li>
    <li><code>(signal-catch-sigint [procedure])</code>: <code>bool</code> The procedure passed to <code>signal-catch-sigint</code> should take no arguments and will be run upon sigint being received. This will override all default behavior (if you still want the program to exit you will need to call <code>(exit)</code> from within the procedure). This procedure (<code>signal-catch-sigint</code>) should be called at the root level of your program as it will only have access to the global environment.</li>
    <li><code>(mod-path)</code>: <code>string</code></li>
  </ul>
</details>


#### OS Related

`path-exists?`, `path-is-dir?`, `path-abs`, `path-join`, `path-extension`, `path-base`, `path-dir`, `path-glob`, `chmod`, `chdir`, `env`, `subprocess`, `mkdir`, `rm`, `mv`, `pwd`


<details>
  <summary>OS API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(path-exists? [filepath: string])</code>: <code>bool</code></li>
    <li><code>(path-is-dir? [filepath: string])</code>: <code>bool</code></li>
    <li><code>(path-abs [filepath: string])</code>: <code>string</code></li>
    <li><code>(path-join [string])</code>: <code>string</code></li>
    <li><code>(path-extension [filepath: string] [[replacement-extension: string]])</code>: <code>string</code></li>
    <li><code>(path-base [filepath: string])</code>: <code>string</code></li>
    <li><code>(path-dir [filepath: string])</code>: <code>string</code></li>
    <li><code>(path-glob [filepath: string])</code>: <code>list</code></li>
    <li><code>(chmod [filepath: string] [file-mode: number])</code>: <code>()</code></li>
    <li><code>(chdir [filepath: string])</code>: <code>()</code></li>
    <li><code>(env [[env-key: string]] [[env-value: string]])</code>: <code>list|string|()</code></li>
    <li><code>(subprocess [list] [[output-redirection: IOHandle|#f]] [[error-redirection: IOHandle|#f]] [[input-redirection: IOHandle|#f]])</code>: <code>number</code> (Passing `#f` to output redirection allows you to do stdout while still redirecting stderr. The `#f` for stderr only exists for symetry)</li>
    <li><code>(mkdir [path: string] [permissions: number] [[make-all: bool]])</code>: <code>()</code></li>
    <li><code>(rm [path: string] [[make-all: bool]])</code>: <code>()</code></li>
    <li><code>(mv [from-path: string] [to-path: string])</code>: <code>()</code></li>
    <li><code>(pwd)</code>: <code>path: string</code></li>
  </ul>
</details>

#### Net Related

Implemented:

`net-conn`, `url-scheme`, `url-host`, `url-port`, `url-path`, `url-query`, `hostname`

_Note_: For the time being, `net-conn` can be used to make most requests. At the moment the tls settings are not particularly secure and I am looking into ways to allow for more granular customization of both the timeout and the tls without weighing down the function or introducing more types.


<details>
  <summary>Net API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(net-conn [host: string] [port: string|number] [[use-tls: bool]] [[timeout-seconds: number]])</code>: <code>IOHandle</code></li>
    <li><code>(url-scheme [url: string] [[new-scheme: string]])</code>: <code>string</code></li>
    <li><code>(url-host [url: string] [[new-host: string]])</code>: <code>string</code></li>
    <li><code>(url-port [url: string] [[new-port: string]])</code>: <code>string</code></li>
    <li><code>(url-path [url: string] [[new-path: string]])</code>: <code>string</code></li>
    <li><code>(url-query [url: string] [[new-query: string]])</code>: <code>string</code></li>
    <li><code>(hostname)</code>: <code>string</code></li>
  </ul>
</details>

#### Date Time Related

Implemented:

`timestamp`, `date`, `timestamp->date`, `date->timestamp`, `date-format`, `sleep`, `date-default-format`

<details>
  <summary>Date/Time API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
    <li><code>(timestamp)</code>: <code>timestamp: number</code></li>
    <li><code>(date)</code>: <code>date: string</code></li>
    <li><code>(timestamp->date [timestamp: string|number] [[format: string]])</code>: <code>string</code></li>
    <li><code>(date->timestamp [date: string] [format: string])</code>: <code>timestamp: number</code></li>
    <li><code>(date-format [start-format: string] [date: string] [end-format: string])</code>: <code>string</code></li>
    <li><code>(sleep [[milliseconds: number]])</code>: <code>()</code></li>
    <li><code>(date-default-format)</code>: <code>string</code> (returns a format string for the default format produced by <code>date</code>)</li>
  </ul>
  <p>In the above procedures there are strings listed as "format". The format string is similar to the format string used in PHP, Python, etc.</p>
  <p>Certain characters are escaped by <code>%</code> symbols followed by a single character. Here is the list of the conversions:</p>
  <dl>
    <dt><code>%a</code></dt><dd>12 hour segment, lowercase: <code>pm</code></dd>
    <dt><code>%A</code></dt><dd>12 hour segment, uppercase: <code>PM</code></dd>
    <dt><code>%d</code></dt><dd><b>Day</b> without leading zero: <code>4</code></dd>
    <dt><code>%D</code></dt><dd><b>Day</b> with leading zero: <code>04</code></dd>
    <dt><code>%e</code></dt><dd><b>Day</b> without leading zero, but with space padding: <code>  4</code></dd>
    <dt><code>%f</code></dt><dd><b>Month</b> name, short: <code>Jan</code></dd>
    <dt><code>%F</code></dt><dd><b>Month</b> name, long: <code>January</code></dd>
    <dt><code>%g</code> or <code>%G</code></dt><dd><b>Hour</b>, 24 hour format: <code>15</code></dd>
    <dt><code>%h</code></dt><dd><b>Hour</b>, 12 hour format without leading zero: <code>2</code></dd>
    <dt><code>%H</code></dt><dd><b>Hour</b>, 12 hour format with leading zero: <code>02</code></dd>
    <dt><code>%i</code></dt><dd><b>Minutes</b> without leading zero: <code>4</code></dd>
    <dt><code>%I</code></dt><dd><b>Minutes</b> with leading zero: <code>04</code></dd>
    <dt><code>%m</code></dt><dd><b>Month number</b>, without leading zero: <code>6</code></dd>
    <dt><code>%M</code></dt><dd><b>Month number</b>, with leading zero: <code>06</code></dd>
    <dt><code>%o</code></dt><dd>Timezone <b>offset</b>, without minutes: <code>-07</code></dd>
    <dt><code>%O</code></dt><dd>Timezone <b>offset</b>, with minutes: <code>-0700</code></dd>
    <dt><code>%s</code></dt><dd><b>Seconds</b>, without leading zero: <code>5</code></dd>
    <dt><code>%S</code></dt><dd><b>Seconds</b>, with leading zero: <code>05</code></dd>
    <dt><code>%w</code></dt><dd>Short <b>weekeday</b>: <code>Wed</code></dd>
    <dt><code>%W</code></dt><dd>Long <b>weekday</b>: <code>Wednesday</code></dd>
    <dt><code>%y</code></dt><ddaTwo digit <b>year</b>: <code>21</code></dd>
    <dt><code>%Y</code></dt><dd>Four digit <b>year</b>: <code>2021</code></dd>
    <dt><code>%Z</code></dt><dd><b>Time zone</b>, as three chars: <code>UTC</code>, <code>PST</code>, etc.</dd>
    <dt><code>%</code></dt><dd>Will yield a literal <code>%</code></dd>
    <dt><code>anything else</code></dt><dd>A percent followed by an unrecognized char will yield a <code>?</code>, as will a hanging % as the last char in the string</dd>
  </dl>
  <p>So the string: <code>"%F %e, %Y %h:%I%a"</code> would be equivalend to something like: <code>"August  8, 2021 4:03pm"</code></p>
  <p>When a string is converted to a date and no time zone information is present, slope will assume the user's local time.</p>
</details>

#### GUI Related

Slope offers a limited wrapper around the [fyne](https://fyne.io) gui toolkit. This is a fully _optional_ build item. See the [build instructions](#building-from-source) for information about building, or not, the gui module.

The module does not offer any free/object drawing or animation. It offers a selection of widgets and containers and some ability to manipulate them and use callbacks from them. This allows for many practical applications that are based more around form fields than graphics.

Implemented:

`dialog-error`, `dialog-info`, `dialog-open-file`, `dialog-save-file`, `container`, `container-add-to`, `conainer-scroll`, `container-size`, `gui-set-char-handlers`, `gui-add-shortcut`, `gui-add-window`, `gui-create`, `gui-list-windows`, `gui-use-light-theme`, `widget-add-to-size`, `widget-disable`, `widget-get-text`, `widget-hide`, `widget-make-button`, `widget-make-checkbox`, `widget-make-entery`, `widget-make-hyperlink`, `widget-make-image`, `widget-make-label`, `widget-make-markdown`, `widget-make-multiline`, `widget-make-password`, `widget-make-select`, `widget-make-separator`, `widget-make-spacer`, `widget-set-text`, `widget-resize`, `widget-show`, `widget-size`, `window-allow-resize`, `window-center`, `window-close`, `window-hide`, `window-resize`,`window-set-content`, `window-set-fullscreen`, `window-set-title`, `window-show`, `window-show-and-run`

<details>
  <summary>GUI API Description</summary>
  <p>Anything inside `[]` is a placeholder representing an argument. Anything inside `[[]]` is a placeholder representing an _optional_ argument. `...` indicates a variadic value.</p>
  <p>What the procedure returns is indicated, though any procedure can return an exception in addition to the given type.</p>
  <ul>
  <li><code>(dialog-error [gui-root] [window-name: string] [error: string]) => ()</code></li>
  <li><code>(dialog-info [gui-root] [window-name: string] [title: string] [message: string]) => ()</code></li>
  <li><code>(dialog-open-file [app: gui-root] [window-name: string] [callback: lambda]) => ()</code></li>
  <li><code>(dialog-save-file [app: gui-root] [window-name: string] [callback: lambda]) => ()</code></li>
  <li><code>(container [layout: string] [[columns: number]] [contents: gui-widget|container...])</code><p>Container is a tricky signature with a few different patterns. Check `usage` for more details as this does not give the full picture</p></li>
  <li><code>(container-add-to [parent: gui-container] [child: gui-widget|gui-container]) => #t</code></li>
  <li><code>(container-scroll [gui-container]) => gui-container</code></li>
  <li><code>(container-size [gui-container]) => list</code></li>
  <li><code>(gui-add-shortcut [gui-root] [window-name: string] [callback: lambda] [key: string] [modifier: string]) => #t</code></li>
  <li><code>(gui-add-window [app: gui-root] [name: string])</code></li>
  <li><code>(gui-create) => gui-root</code></li>
  <li><code>(gui-list-windows [app: gui-root]) => list</code></li>
  <li><code>(gui-set-char-handlers [gui-root] [window-name: string] [callback: lambda]) => #t</code></li>
  <li><code>(gui-use-light-theme [[bool]]) => bool</code></li>
  <li><code>(widget-add-to-size [gui-widget] [width-change: number] [height-change: number]) => gui-widget</code></li>
  <li><code>(widget-disable [gui-widget] [[set-disabled: bool]]) => bool</code></li>
  <li><code>(widget-get-text [gui-widget]) => string</code></li>
  <li><code>(widget-hide [gui-widget]) => #t</code></li>
  <li><code>(widget-make-button [text: string] [callback: lambda] [[alignment: number]] [[importance: number]]) => gui-widget</code></li>
  <li><code>(widget-make-checkbox [text: string] [callback: lambda]) => gui-widget</code></li>
  <li><code>(widget-make-entry [placeholder-text: string] [[wrap-text: bool]] [[validation-callback: lambda]]) => gui-widget</code></li>
  <li><code>(widget-make-hyperlink [text: string] [url: string]) => gui-widget</code></li>
  <li><code>(widget-make-image [source: string] [location: string] [[mode: string]] [[min-width: number]] [[min-height]]) => gui-widget</code></li>
  <li><code>(widget-make-label [text: string] [[alignment: number]] [[wrap-text: bool]]) => gui-widget</code></li>
  <li><code>(widget-make-markdown [markdown-text: string] [[wrap-text: bool]]) => gui-widget</code></li>
  <li><code>(widget-make-multiline [placeholder-text: string] [[wrap-text: bool]] [[use-monospace-font: bool]] [[validation-callback: lambda]]) => gui-widget</code></li>
  <li><code>(widget-make-password [placeholder-text: string] [[wrap-text: bool]] [[validation-callback: lambda]]) => gui-widget</code></li>
  <li><code>(widget-make-select [options: list] [callback: lambda] [[alignment: number]]) => gui-widget</code></li>
  <li><code>(widget-make-separator) => gui-widget</code></li>
  <li><code>(widget-make-spacer) => gui-widget</code></li>
  <li><code>(widget-set-text [gui-widget] [string]) => string</code></li>
  <li><code>(widget-resize [gui-widget] [width: number] [height: number]) => gui-widget</code></li>
  <li><code>(widget-show [gui-widget]) => #t</code></li>
  <li><code>(widget-size [gui-widget]) => list</code></li>
  <li><code>(window-allow-resize [app: gui-root] [window-name: string] [bool]) => #t</code></li>
  <li><code>(window-center [app: gui-root] [window-name: string]) => #t</code></li>
  <li><code>(window-close [app: gui-root] [window-name: string]) => #t</code></li>
  <li><code>(window-hide [app: gui-root] [window-name: string]) => #t</code></li>
  <li><code>(window-resize [app: gui-root] [window-name: string] [width: number] [height: number]) => #t</code></li>
  <li><code>(window-set-content [app: gui-root] [window-name: string] [content: gui-widget|gui-container]) => #t</code></li>
  <li><code>(window-set-fullscreen [app: gui-root] [window-name: string] [set-to-full-screen: bool]) => #t</code></li>
  <li><code>(window-set-title [app: gui-root] [window-name: string] [title: string]) => #t</code></li>
  <li><code>(window-show [app: gui-root] [window-name: string]) => #t</code></li>
  <li><code>(window-show-and-run [app: gui-root] [window-name: string]) => #t</code></li>
  </ul>
</details>

![Demonstration of slope gui](/images/slope-gui-demo.png)

---


\*the general write, read, and close operations in the IO section apply to net objects as well as files and string buffers

## Preload

slope allows for files in a specific directory to be automatically loaded. For this to occur slope should be invoked with the `-L` flag. When that flag is passed slope will do its normal setup routine then preload all files in the preload folder then go on to whatever its main task is (repl, run file, run single line from command).

slope will use the variable `$SLOPE_PRELOAD_DIR` if it is defined and not empty. Otherwise, if `$XDG_DATA_HOME` is defined and not empty it will use `$XDG_DATA_HOME/slope/preload/`. Lastly it will use `~/.local/share/slope/preload/`.

Unlike modules, arbitrary code execution _can_ occur with preloads (not just `define` expressions). The rationalle is that you, the user, have put this code there and as such it is considered safe and in your control. Preloads are a really easy way to make procedures and variables outside of the standard library, but that you regularly use, available at a repl session without having to manually load them.


## Modules

slope has a basic module system. When you use the `load` procedure you are loading a single file from a given path. Anything in that file will be run when it is loaded. Using `load-mod` allows you to load a module from a designated location on your system by simply passing the module's name (as represented by the directory name it is contained within).

`load-mod` will load modules from the first non-empty-string value it finds out of these options:

1. `$SLOPE_MOD_PATH`
2. `$XDG_DATA_HOME/slope/modules`
3. `~/.local/share/slope/modules`
4. `/usr/local/lib/slope/modules`

So, a hypethetical module named `test` would be found at `~/.local/share/slope/modules/test`.

Note that the fourth item, above, is the global (system-wide) module directory. `slp` can install modules to this directory by passing the `-g` or `--global` flag while acting as a user that has access to writing files in the `/usr/local` heirarchy. You may, alternately, clone modules directly to this path. A local module will always be used before a global module.


### Package Management

slope has a package manager, `slp`, available at [https://git.rawtext.club/slope-lang/slp](https://git.rawtext.club/slope-lang/slp). Once installed you will be able to search/browse packages, install, remove, and update modules. You can also use `slp` to generate new module skeletons, read docs, etc. The `slp` repository has information on how to get modules added to the package registry. This is, at present, the best way to deal with modules and is highly recommended but by no means required. The same thing can be done by finding a repository with a module and cloning it to your module path. So long as it is a valid module it will be loadable with `load-mod`.


### How Modules are Processed

When `load-mod` is called the module will be located and `main.slo` will be parsed. All expressions are ignored except `define`, `load-mod`, and `load-mod-file`. Arbitrary code cannot be run (there are no side effects to `load-mod` or `load-mod-file` other than populating the global environment with the defined values from the given module).

A `define` expression _will still be fully evaluated_. As such, something like: `(define twenty (+ 15 5))` is perfectly fine. But something like `(display "I am in a module")` will not be evaluated.

Calling `load-mod` from within a module allows modules to have their own dependencies. Calling `load-mod-file` allows modules to be made up of more files than just `main.slo`. More info about files can be found in the next section.

### Module Documentation

Modules may (one could argue _should_) contain a value named `_USAGE` in their `main.slo` file. This value should be an association list detailing each procedure. It should take the form of:

``` scheme
; This is for a module named: "mymodule"

(define _USAGE [
  ["my-proc" "(my-proc [name: string]) => bool\nFurther info can go here. Provide what would be useful"]
  ["my-other-proc" "(my-other-proc [name: string] [count: number]) => string"]
])
```

This value is special and will be parsed so that the information can be retrieved at the repl:

``` scheme
; to get a list of procedures within the given module
(usage "mymodule" #f) ; the #f is so that the interpreter knows that you are requesting a module listing and not trying to get usage for a builtin named "mymodule", an empty string can be used in place of #f if preferred

; to get info on a single procedure
(usage "mymodule" "my-proc")
```

It is good form to always include the procedure signature as the first item in the usage instructions. Arguments are included in the format: `[name: type]`. Optional arguments use double square brackets: `[[name: type]]`. The output should be denoted after the close of the function signature via `=> type`. If multiple types can be used or returned use a `|`: `(my-func [input: number|string]) => procedure|#f`. If a specific value can be returned, you can include that, as `#f` was used in the previous example. Many procedures return some value _or_ false (`#f`) to indicate a failure.

### Module Files

#### main.slo

Taking our above example of a module named `test`, the bare minimum that the folder `~/.local/share/slope/modules/test` should contain is a file named `main.slo`. When a module is loaded, this is the file that is looked for. If it is not present, loading will fail and a runtime panic will result.

#### Additional source files

You may, optionally, have other code files present in the module and can load them by calling `load-mod-file`. This procedure should be given a path relative to the module root directory. So within our `main.slo` file we might call something like: `(load-mod-file "./submodules/unit-tests.slo")`. Something to note is that a call to `load-mod-file` cannot jump out of the module's root directory (doing something like `../../../some-other-code.py` or `/etc/something` will result in a runtime panic).

#### module.json

Modules found in the package registry will include this file. It contains metadata related to the package including things like description, version, repository url, homepage, author, and, most importantly the dependency list (so that a modules dependencies can also be installed).

If you are making a module and do not know what should be here you can look at an existing module to get a sense of it. Or, with `slp` installed, you can just run `slp gen` and it will build out your module skeleton for you.



## Building from source

A Makefile is available. Running `make` will build in the local directory. By default, this will _not_ include the gui module. To include the gui module in the build run `make gui`. Running `sudo make install` after running `make` or `make gui` will install the software globally (sudo may or may not be needed depending on your setup, or the functionality may be provided by a different tool such as doas).  _slope_ was written with Go 1.16, but will likely build on older versions with limited modifications.

``` sh
git clone https://git.rawtext.club/slope-lang/slope
cd slope
make       # or `make gui`

# For global install run the following
# after running `make` or `make gui`
sudo make install
```

If you prefer to build directly with the go compiler rather than the provided makefile you can do a basic build with one of two commands (depending on whether or not you want gui):

``` sh
# WITHOUT gui
go build

# WITH gui
go build -tags gui
```

The above build methods, without the makefile, will not provide a version hash to the program, but will successfully build the program. The makefile has some nice features and is definitely the recommended route.

If you run into issues getting the gui version to build, it may be because of slope's dependency on the [fyne](https://github.com/fyne-io/fyne) toolkit. Their issues or build instructions may be helpful. That said, gui is presently an experimental feature and has been made an optional component so that an upstream component will not block installs of the interpreter.

## Running

All of the shell commands in this section assume slope is on your `$PATH`.


### The REPL

To run the slope REPL environment execute the following at your terminal prompt:

``` sh
slope
```

You can, at your option, use the `-L` flag to preload the environment with your [preload](#preload) content.

From the REPL you can execute `(license)` to view the license. To exit, you execute `(exit)`.

While in the REPL readline compatible key commands are available. For a list of compatible key commands please see [this list](https://github.com/peterh/liner#line-editing).

In addition to key commands, the repl supports command completion for any identifier that has `usage` information available. Since `usage` supports getting usage information from _loaded_ (via `load-mod`) modules, any usage information provided by said modules will be included in completion. To complete a known identifier press the `Tab` key. The first found completion will appear. Subsequent presses of the `Tab` key will cycle through any other completion candidates that were found.

### One Liners

To run a one liner execute the following at your terminal prompt:

``` sh
slope -run "(display (+ 1 3))"
```

You may, of course, replace the contents between the quotes with your own one liner. In general all features of slope are available from a one liner, including loading modules.

You can, at your option, use the `-L` flag to preload the environment with your [preload](#preload) content.


### Slope Files

To run a file execute slope with the file path (an example path has been provided below):

``` sh
slope ./myfile.slo
```


### Other Options

To see command line options:

``` sh
slope -h
```


To load all slope file in your [preload](#preload) folder before proceeding to run the file, one liner, or repl:

``` sh
slope -L
```


To see version information:

``` sh
slope -v
```


To install a local module:

``` sh
slope -install /path/to/module
```
The filepath should lead to a folder that contains at the very least a `main.slo` file. If installing the module would overwrite an existing module on the module path you will be warned and need to give consent to overwrite the module.


To turn on interpreter runtime stack traces:

``` sh
slope -debug
```



## License

slope is available under the terms of the MIT license. See either the `LICENSE` file in this repository or run `(license)` from slope (via file, `-run`, or REPL).

